# Copyright (c) 2011-2017, Dan Crosta
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from flask import abort
from pymongo import collection
from pymongo import database
from pymongo import mongo_client


class MongoClient(mongo_client.MongoClient):

    """Wrapper for :class:`~pymongo.mongo_client.MongoClient`.

    Returns instances of Flask-PyMongo
    :class:`~flask_pymongo.wrappers.Database` instead of native PyMongo
    :class:`~pymongo.database.Database` when accessed with dot notation.

    """

    def __getattr__(self, name):  # noqa: D105
        attr = super(MongoClient, self).__getattr__(name)
        if isinstance(attr, database.Database):
            return Database(self, name)
        return attr

    def __getitem__(self, item):  # noqa: D105
        attr = super(MongoClient, self).__getitem__(item)
        if isinstance(attr, database.Database):
            return Database(self, item)
        return attr


class Database(database.Database):

    """Wrapper for :class:`~pymongo.database.Database`.

    Returns instances of Flask-PyMongo
    :class:`~flask_pymongo.wrappers.Collection` instead of native PyMongo
    :class:`~pymongo.collection.Collection` when accessed with dot notation.

    """

    def __getattr__(self, name):  # noqa: D105
        attr = super(Database, self).__getattr__(name)
        if isinstance(attr, collection.Collection):
            return Collection(self, name)
        return attr

    def __getitem__(self, item):  # noqa: D105
        item_ = super(Database, self).__getitem__(item)
        if isinstance(item_, collection.Collection):
            return Collection(self, item)
        return item_


class Collection(collection.Collection):

    """Sub-class of PyMongo :class:`~pymongo.collection.Collection` with helpers.

    """

    def __getattr__(self, name):  # noqa: D105
        attr = super(Collection, self).__getattr__(name)
        if isinstance(attr, collection.Collection):
            db = self._Collection__database
            return Collection(db, attr.name)
        return attr

    def __getitem__(self, item):  # noqa: D105
        item_ = super(Collection, self).__getitem__(item)
        if isinstance(item_, collection.Collection):
            db = self._Collection__database
            return Collection(db, item_.name)
        return item_

    def find_one_or_404(self, *args, **kwargs):
        """Find a single document or raise a 404.

        This is like :meth:`~pymongo.collection.Collection.find_one`, but
        rather than returning ``None``, cause a 404 Not Found HTTP status
        on the request.

        .. code-block:: python

            @app.route("/user/<username>")
            def user_profile(username):
                user = mongo.db.users.find_one_or_404({"_id": username})
                return render_template("user.html",
                    user=user)

        """
        found = self.find_one(*args, **kwargs)
        if found is None:
            abort(404)
        return found



from contextlib import contextmanager
import time

import pymongo
import pytest

from flask_pymongo.tests.util import FlaskRequestTest
import flask_pymongo


class CouldNotConnect(Exception):
    pass


@contextmanager
def doesnt_raise(exc=BaseException):
    try:
        yield
    except exc:
        pytest.fail("{} was raised but should not have been".format(exc))


class FlaskPyMongoConfigTest(FlaskRequestTest):

    def setUp(self):
        super(FlaskPyMongoConfigTest, self).setUp()

        conn = pymongo.MongoClient(port=self.port)
        conn.test.command("ping")  # wait for server

    def tearDown(self):
        super(FlaskPyMongoConfigTest, self).tearDown()

        conn = pymongo.MongoClient(port=self.port)

        conn.drop_database(self.dbname)
        conn.drop_database(self.dbname + "2")

    def test_config_with_uri_in_flask_conf_var(self):
        uri = "mongodb://localhost:{}/{}".format(self.port, self.dbname)
        self.app.config["MONGO_URI"] = uri

        mongo = flask_pymongo.PyMongo(self.app, connect=True)

        _wait_until_connected(mongo)
        assert mongo.db.name == self.dbname
        assert ("localhost", self.port) == mongo.cx.address

    def test_config_with_uri_passed_directly(self):
        uri = "mongodb://localhost:{}/{}".format(self.port, self.dbname)

        mongo = flask_pymongo.PyMongo(self.app, uri, connect=True)

        _wait_until_connected(mongo)
        assert mongo.db.name == self.dbname
        assert ("localhost", self.port) == mongo.cx.address

    def test_it_fails_with_no_uri(self):
        self.app.config.pop("MONGO_URI", None)

        with pytest.raises(ValueError):
            flask_pymongo.PyMongo(self.app)

    def test_multiple_pymongos(self):
        uri1 = "mongodb://localhost:{}/{}".format(self.port, self.dbname)
        uri2 = "mongodb://localhost:{}/{}".format(self.port, self.dbname + "2")

        mongo1 = flask_pymongo.PyMongo(self.app, uri1)  # noqa: F841 unused variable
        mongo2 = flask_pymongo.PyMongo(self.app, uri2)  # noqa: F841 unused variable

        # this test passes if it raises no exceptions

    def test_custom_document_class(self):
        class CustomDict(dict):
            pass

        uri = "mongodb://localhost:{}/{}".format(self.port, self.dbname)
        mongo = flask_pymongo.PyMongo(self.app, uri, document_class=CustomDict)
        assert mongo.db.things.find_one() is None, "precondition failed"

        mongo.db.things.insert_one({"_id": "thing", "val": "foo"})

        assert type(mongo.db.things.find_one()) == CustomDict

    def test_it_doesnt_connect_by_default(self):
        uri = "mongodb://localhost:{}/{}".format(self.port, self.dbname)

        mongo = flask_pymongo.PyMongo(self.app, uri)

        with pytest.raises(CouldNotConnect):
            _wait_until_connected(mongo, timeout=0.2)

    def test_it_doesnt_require_db_name_in_uri(self):
        uri = "mongodb://localhost:{}".format(self.port)

        with doesnt_raise(Exception):
            mongo = flask_pymongo.PyMongo(self.app, uri)

        assert mongo.db is None


def _wait_until_connected(mongo, timeout=1.0):
    start = time.time()
    while time.time() < (start + timeout):
        if mongo.cx.nodes:
            return
        time.sleep(0.05)
    raise CouldNotConnect("could not prove mongodb connected in %r seconds" % timeout)



from hashlib import md5
from io import BytesIO

from bson.objectid import ObjectId
from gridfs import GridFS
from werkzeug.exceptions import NotFound
import pytest

from flask_pymongo.tests.util import FlaskPyMongoTest


class GridFSCleanupMixin(object):

    def tearDown(self):
        gridfs = GridFS(self.mongo.db)
        files = list(gridfs.find())
        for gridfile in files:
            gridfs.delete(gridfile._id)

        super(GridFSCleanupMixin, self).tearDown()


class TestSaveFile(GridFSCleanupMixin, FlaskPyMongoTest):

    def test_it_saves_files(self):
        fileobj = BytesIO(b"these are the bytes")

        self.mongo.save_file("my-file", fileobj)

        gridfs = GridFS(self.mongo.db)
        assert gridfs.exists({"filename": "my-file"})

    def test_it_guesses_type_from_filename(self):
        fileobj = BytesIO(b"these are the bytes")

        self.mongo.save_file("my-file.txt", fileobj)

        gridfs = GridFS(self.mongo.db)
        gridfile = gridfs.find_one({"filename": "my-file.txt"})
        assert gridfile.content_type == "text/plain"

    def test_it_saves_files_with_props(self):
        fileobj = BytesIO(b"these are the bytes")

        self.mongo.save_file("my-file", fileobj, foo="bar")

        gridfs = GridFS(self.mongo.db)
        gridfile = gridfs.find_one({"filename": "my-file"})
        assert gridfile.foo == "bar"

    def test_it_returns_id(self):
        fileobj = BytesIO(b"these are the bytes")

        _id = self.mongo.save_file("my-file", fileobj, foo="bar")

        assert type(_id) is ObjectId


class TestSendFile(GridFSCleanupMixin, FlaskPyMongoTest):

    def setUp(self):
        super(TestSendFile, self).setUp()

        # make it bigger than 1 gridfs chunk
        self.myfile = BytesIO(b"a" * 500 * 1024)
        self.mongo.save_file("myfile.txt", self.myfile)

    def test_it_404s_for_missing_files(self):
        with pytest.raises(NotFound):
            self.mongo.send_file("no-such-file.txt")

    def test_it_sets_content_type(self):
        resp = self.mongo.send_file("myfile.txt")
        assert resp.content_type.startswith("text/plain")

    def test_it_sets_content_length(self):
        resp = self.mongo.send_file("myfile.txt")
        assert resp.content_length == len(self.myfile.getvalue())

    def test_it_sets_supports_conditional_gets(self):
        # a basic conditional GET
        environ_args = {
            "method": "GET",
            "headers": {
                "If-None-Match": md5(self.myfile.getvalue()).hexdigest(),
            },
        }

        with self.app.test_request_context(**environ_args):
            resp = self.mongo.send_file("myfile.txt")
            assert resp.status_code == 304

    def test_it_sets_cache_headers(self):
        resp = self.mongo.send_file("myfile.txt", cache_for=60)
        assert resp.cache_control.max_age == 60
        assert resp.cache_control.public is True

    def test_it_streams_results(self):
        resp = self.mongo.send_file("myfile.txt")
        assert resp.is_streamed
